from dataclasses import dataclass, field

import numpy as np
from dyce import H
from numpy.typing import ArrayLike

from .core import BaseDice


@dataclass(slots=True)
class DiceAdd(BaseDice):
    items: tuple[BaseDice, ...]

    def histogram(self) -> H:
        return sum(i.histogram() for i in self.items)  # type: ignore

    def __str__(self) -> str:
        return "(" + " + ".join(str(i) for i in self.items) + ")"

    def max(self) -> int:
        return np.sum([i.max() for i in self.items])

    def min(self) -> int:
        return np.sum([i.min() for i in self.items])

    def generate(self, items: int) -> ArrayLike:
        res = np.zeros(items, dtype=np.int_)
        for i in self.items:
            res += i.generate(items)  # type: ignore
        return res


@dataclass(slots=True)
class DiceSub(BaseDice):
    items: tuple[BaseDice, ...]

    def histogram(self) -> H:
        result = self.items[0].histogram()
        for i in self.items[1:]:
            result -= i.histogram()  # type: ignore
        return result  # type: ignore

    def __str__(self) -> str:
        return "(" + " - ".join(str(i) for i in self.items) + ")"

    def max(self) -> int:
        # The maximum value for subtraction is calculated by subtracting the min values of all but the first dice from the max value of the first dice.
        max_first_item = self.items[0].max()
        sum_of_max_of_others = np.sum([i.min() for i in self.items[1:]])
        return max_first_item - sum_of_max_of_others

    def min(self) -> int:
        # The minimum value for subtraction is calculated by subtracting the max values of all but the first dice from the min value of the first dice.
        min_first_item = self.items[0].min()
        sum_of_max_of_others = np.sum([i.max() for i in self.items[1:]])
        min_value = min_first_item - sum_of_max_of_others
        return min_value

    def generate(self, items: int) -> ArrayLike:
        result = self.items[0].generate(items)
        for item in self.items[1:]:
            result -= item.generate(items)  # type: ignore
        return result


@dataclass(slots=True)
class DiceMul(BaseDice):
    items: tuple[BaseDice, ...]

    def histogram(self) -> H:
        result = self.items[0].histogram()
        for i in self.items[1:]:
            result *= i.histogram()  # type: ignore
        return result  # type: ignore

    def __str__(self) -> str:
        return "(" + " * ".join(str(i) for i in self.items) + ")"

    def max(self) -> int:
        # Calculate the maximum possible outcome by multiplying the maximum values of all included dice.
        return int(np.prod([i.max() for i in self.items]))

    def min(self) -> int:
        # Calculate the minimum possible outcome by multiplying the minimum values of all included dice.
        return int(np.prod([i.min() for i in self.items]))

    def generate(self, items: int) -> ArrayLike:
        res = np.ones(items, dtype=np.int_)
        for i in self.items:
            res *= i.generate(items)  # type: ignore
        return res


@dataclass(slots=True)
class DiceDiv(BaseDice):
    items: tuple[BaseDice, ...]

    def histogram(self) -> H:
        result = self.items[0].histogram()
        for i in self.items[1:]:
            result /= i.histogram()  # type: ignore
        return result  # type: ignore

    def __str__(self) -> str:
        return "(" + " / ".join(str(i) for i in self.items) + ")"

    def max(self) -> int:
        # Divide the max value of the first item by the product of min values of the rest.
        numerator = self.items[0].max()
        # Calculate the product of min values for the rest of the items.
        denominators_product = np.prod([i.min() for i in self.items[1:]])
        # Perform division, ensuring no division by zero.
        return int(numerator // denominators_product if denominators_product else 0)

    def min(self) -> int:
        # Divide the min value of the first item by the product of max values of the rest.
        numerator = self.items[0].min()
        # Calculate the product of max values for the rest of the items.
        denominators_product = np.prod([i.max() for i in self.items[1:]])
        # Perform division, ensuring no division by zero.
        return int(numerator // denominators_product if denominators_product else 0)

    def generate(self, items: int) -> ArrayLike:
        result = self.items[0].generate(items)
        for item in self.items[1:]:
            result //= item.generate(items)  # type: ignore
        return result



from dataclasses import dataclass
from functools import partial
from typing import Protocol

import numpy as np
from dyce import H
from dyce.evaluation import HResult, expandable
from numpy.typing import ArrayLike

from .core import BaseDice
from .misc import DiceModifier, _wrap_scalar


@dataclass(slots=True)
class BaseCompare(BaseDice, Protocol):
    dice: BaseDice
    compare: BaseDice

    @staticmethod
    def _with_cap(roll_values: ArrayLike, cmp_values: ArrayLike) -> tuple[ArrayLike, ArrayLike]: ...

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool: ...

    @staticmethod
    def _modify_input_histogram(dice: H, compare: H) -> tuple[H, H]:
        return dice, compare

    def histogram(self) -> H:
        @expandable
        def cmp(dice: HResult, compare: HResult):
            if not self._compare_histogram_outcome(dice.outcome, compare.outcome):  # type: ignore
                return compare.outcome
            return dice.outcome

        return cmp(*self._modify_input_histogram(self.dice.histogram(), self.compare.histogram()))

    def generate(self, items: int) -> ArrayLike:
        result_rolls = self.dice.generate(items)
        cmp_rolls = self.compare.generate(items)
        return self._with_cap(result_rolls, cmp_rolls)  # type: ignore


@dataclass(slots=True)
class Lt(BaseCompare):
    def __str__(self) -> str:
        return f"{self.dice}<{self.compare}"

    def max(self) -> int:
        return min(self.compare.max() - 1, self.dice.max())

    def min(self) -> int:
        return min(self.dice.min(), self.max())

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice < compare

    @staticmethod
    def _modify_input_histogram(dice: H, compare: H) -> tuple[H, H]:
        return dice, compare - 1  # type: ignore

    @staticmethod
    def _with_cap(roll_values: ArrayLike, cmp_values: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        return np.minimum(roll_values, (cmp_values - 1))  # type: ignore


@dataclass(slots=True)
class Le(BaseCompare):
    def __str__(self) -> str:
        return f"{self.dice}<={self.compare}"

    def max(self) -> int:
        return min(self.compare.max(), self.dice.max())

    def min(self) -> int:
        return min(self.dice.min(), self.max())

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice <= compare

    @staticmethod
    def _with_cap(roll_values: ArrayLike, cmp_values: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        return np.minimum(roll_values, cmp_values)  # type: ignore


@dataclass(slots=True)
class Gt(BaseCompare):
    def __str__(self) -> str:
        return f"{self.dice}>{self.compare}"

    def max(self) -> int:
        return max(self.dice.max(), self.compare.max() + 1)

    def min(self) -> int:
        return max(self.dice.min(), self.compare.min() + 1)

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice > compare

    @staticmethod
    def _modify_input_histogram(dice: H, compare: H) -> tuple[H, H]:
        return dice, compare + 1  # type: ignore

    @staticmethod
    def _with_cap(roll_values: ArrayLike, cmp_values: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        return np.maximum(roll_values, (cmp_values + 1))  # type: ignore


@dataclass(slots=True)
class Ge(BaseCompare):
    def __str__(self) -> str:
        return f"{self.dice}>={self.compare}"

    def max(self) -> int:
        return max(self.dice.max(), self.compare.max())

    def min(self) -> int:
        return max(self.dice.min(), self.compare.min())

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice >= compare

    @staticmethod
    def _with_cap(roll_values: ArrayLike, cmp_values: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        return np.maximum(roll_values, cmp_values)  # type: ignore


class Limit:
    def __gt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(Gt, compare=_wrap_scalar(value))

    def __ge__(self, value: BaseDice | int) -> DiceModifier:
        return partial(Ge, compare=_wrap_scalar(value))

    def __lt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(Lt, compare=_wrap_scalar(value))

    def __le__(self, value: BaseDice | int) -> DiceModifier:
        return partial(Le, compare=_wrap_scalar(value))



from .core import BaseDice, Scalar
from typing import Callable


def _wrap_scalar(value: BaseDice | int) -> BaseDice:
    if isinstance(value, int):
        value = Scalar(value)
    if not isinstance(value, BaseDice):
        raise TypeError("Reroll only support other dices or integers")
    return value


DiceModifier = Callable[[BaseDice], BaseDice]



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from typing import Protocol, runtime_checkable

import numpy as np
from dyce import H
from dyce.evaluation import expandable
from numpy.typing import ArrayLike

from .random import Rng


@runtime_checkable
class BaseDice(Protocol):

    # Interface methods

    def generate(self, items: int) -> ArrayLike: ...

    def max(self) -> int: ...

    def min(self) -> int: ...

    def __str__(self) -> str:
        return super().__str__()

    def histogram(self) -> H: ...

    # Base roll

    def roll(self) -> int:
        return np.sum(self.generate(1))

    # Modifiers

    def kh(self, keep: BaseDice | int = 1) -> BaseDice:
        from .transformations import KeepHighest

        return KeepHighest(self, keep=keep)  # type: ignore

    def kl(self, keep: BaseDice | int = 1) -> BaseDice:
        from .transformations import KeepLowest

        return KeepLowest(self, keep=keep)  # type: ignore

    def dh(self, drop: BaseDice | int = 1) -> BaseDice:
        from .transformations import DropHighest

        return DropHighest(self, drop=drop)  # type: ignore

    def dl(self, drop: BaseDice | int = 1) -> BaseDice:
        from .transformations import DropLowest

        return DropLowest(self, drop=drop)  # type: ignore

    @property
    def r(self):
        return self.reroll()

    def reroll(self, reroll_limit: int = 1):
        from .reroll import _RerollFactory

        return _RerollFactory(dice=self, reroll_limit=reroll_limit)

    @property
    def x(self):
        return self.explode()

    def explode(self, explode_depth: int = 100):
        from .explode import _ExplodeFactory

        return _ExplodeFactory(dice=self, explode_depth=explode_depth)

    # Magic

    def __matmul__(self, other):
        if isinstance(other, int):
            other = Scalar(other)

        if not isinstance(other, BaseDice):
            raise TypeError("Can only add other dices or integers")

        return DiceMany(total=self, dice=other)

    def __rmatmul__(self, other):
        if isinstance(other, int):
            other = Scalar(other)

        if not isinstance(other, BaseDice):
            raise TypeError("Can only add other dices or integers")

        return DiceMany(total=other, dice=self)

    def __add__(self, other):
        if isinstance(other, int):
            other = Scalar(other)

        if not isinstance(other, BaseDice):
            raise TypeError("Can only add other dices or integers")

        from .math import DiceAdd

        items = other.items if isinstance(other, DiceAdd) else [other]
        if isinstance(self, DiceAdd):
            return DiceAdd((*self.items, *items))

        return DiceAdd((self, *items))

    def __radd__(self, other):
        from .math import DiceAdd

        if isinstance(other, int):
            return DiceAdd((Scalar(other), self))

        if not isinstance(other, BaseDice):
            raise TypeError("Can only add other dices or integers")

    def __sub__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only sub other dices or integers")

        from .math import DiceSub

        items = other.items if isinstance(other, DiceSub) else [other]
        if isinstance(self, DiceSub):
            return DiceSub((*self.items, *items))
        return DiceSub((self, *items))

    def __rsub__(self, other):
        from .math import DiceSub

        if isinstance(other, int):
            return DiceSub((Scalar(other), self))
        if not isinstance(other, BaseDice):
            raise TypeError("Can only sub other dices or integers")

    def __mul__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only mul other dices or integers")

        from .math import DiceMul

        items = other.items if isinstance(other, DiceMul) else [other]
        if isinstance(self, DiceMul):
            return DiceMul((*self.items, *items))
        return DiceMul((self, *items))

    def __rmul__(self, other):
        from .math import DiceMul

        if isinstance(other, int):
            return DiceMul((Scalar(other), self))
        if not isinstance(other, BaseDice):
            raise TypeError("Can only mul other dices or integers")

    def __truediv__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

        from .math import DiceDiv

        items = other.items if isinstance(other, DiceDiv) else [other]
        if isinstance(self, DiceDiv):
            return DiceDiv((*self.items, *items))
        return DiceDiv((self, *items))

    def __rtruediv__(self, other):
        from .math import DiceDiv

        if isinstance(other, int):
            return DiceDiv((Scalar(other), self))
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

    __floordiv__ = __truediv__  # type: ignore
    __rfloordiv__ = __rtruediv__  # type: ignore

    def __ge__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

        from .compare import Ge

        return Ge(dice=self, compare=other)  # type: ignore

    def __gt__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

        from .compare import Gt

        return Gt(dice=self, compare=other)  # type: ignore

    def __le__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

        from .compare import Le

        return Le(dice=self, compare=other)  # type: ignore

    def __lt__(self, other):
        if isinstance(other, int):
            other = Scalar(other)
        if not isinstance(other, BaseDice):
            raise TypeError("Can only div other dices or integers")

        from .compare import Lt

        return Lt(dice=self, compare=other)  # type: ignore


@dataclass(slots=True)
class Scalar(BaseDice):
    value: int

    def histogram(self) -> H:
        return H([self.value])  # type: ignore

    def __str__(self) -> str:
        return str(self.value)

    def max(self) -> int:
        return self.value

    def min(self) -> int:
        return self.value

    def generate(self, items: int) -> ArrayLike:
        return np.full(items, self.value)


@dataclass(slots=True)
class Dice(BaseDice):
    sides: int
    minimal: int = field(default=1)

    def histogram(self) -> H:
        return H(range(self.minimal, self.sides + 1))  # type: ignore

    def __str__(self) -> str:
        if self.minimal == 1:
            return f"d{self.sides}"
        return f"d[{self.minimal} to {self.sides}]"

    def max(self) -> int:
        return self.sides

    def min(self) -> int:
        return self.minimal

    def generate(self, items: int) -> ArrayLike:
        return Rng().rng.integers(low=self.minimal, high=self.sides + 1, size=items)


@dataclass(slots=True)
class RangeDice(BaseDice):
    min_value: int
    max_value: int
    step_value: int = field(default=1)

    @cached_property
    def __range(self):
        return range(self.min_value, self.max_value, self.step_value)

    def histogram(self) -> H:
        return H(list(self.__range))  # type: ignore

    def __str__(self) -> str:
        if self.step_value == 1:
            return f"rng({self.min_value},{self.max_value})"
        return f"rng({self.min_value},{self.max_value},{self.step_value})"

    def max(self) -> int:
        # Calculate the number of steps
        num_steps = (self.max_value - self.min_value - 1) // self.step_value + 1
        # Calculate the actual max value in the range
        actual_max = self.min_value + (num_steps - 1) * self.step_value
        return actual_max

    def min(self) -> int:
        return self.min_value

    def generate(self, items: int) -> ArrayLike:
        return Rng().rng.choice(self.__range, size=items, replace=True)


@dataclass(slots=True)
class DiceMany(BaseDice):
    total: BaseDice
    dice: BaseDice

    def histogram(self) -> H:
        many = expandable(lambda total, dice: total.outcome @ dice.h)
        return many(self.total.histogram(), self.dice.histogram())

    def __str__(self) -> str:
        return f"{self.total}{self.dice}"

    def max(self) -> int:
        return self.dice.max() * self.total.max()

    def min(self) -> int:
        return self.dice.min() * self.total.min()

    def generate(self, items: int) -> ArrayLike:
        total_rolls = self.total.generate(items)
        max_rolls = np.max(total_rolls)
        result = np.zeros(items, dtype=np.int_)

        for roll_count in range(1, max_rolls + 1):
            mask = total_rolls >= roll_count  # type: ignore
            num_items_this_round = np.sum(mask)
            # If no items require processing, exit the loop early
            if num_items_this_round == 0:
                break
            # Generate and sum the dice rolls for items requiring them
            result[mask] += self.dice.generate(num_items_this_round)  # type: ignore

        return result



from dataclasses import dataclass, field
from functools import partial
from typing import Protocol

import numpy as np
from dyce import H
from dyce.evaluation import HResult, expandable
from numpy.typing import ArrayLike

from .core import BaseDice
from .misc import DiceModifier, _wrap_scalar


@dataclass(slots=True)
class BaseReroll(BaseDice, Protocol):
    dice: BaseDice
    compare: BaseDice
    reroll_limit: int = field(default=1)

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool: ...

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike: ...

    def max(self) -> int:
        return self.dice.max()

    def min(self) -> int:
        return self.dice.min()

    def histogram(self) -> H:
        dice_hist = self.dice.histogram()

        @expandable(sentinel=dice_hist)
        def _reroll(compare: HResult, dice: HResult):
            if self._compare_histogram_outcome(dice.outcome, compare.outcome):  # type: ignore
                # Replace a roll with rerolled dice
                return _reroll(compare.h, dice.h)  # type: ignore
            else:
                return dice.outcome

        return _reroll(self.compare.histogram(), dice_hist, limit=self.reroll_limit - 1)  # type: ignore

    def generate(self, items: int) -> ArrayLike:
        result = self.dice.generate(items)
        for _ in range(self.reroll_limit):
            compare_values = self.compare.generate(items)
            reroll_mask = self._calculate_reroll_mask(result, compare_values)

            if not np.any(reroll_mask):
                break

            rerolls = self.dice.generate(np.sum(reroll_mask))
            result[reroll_mask] = rerolls  # type: ignore

        return result


@dataclass(slots=True)
class RerollEq(BaseReroll):
    def __str__(self) -> str:
        return f"{self.dice}r{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice == compare

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values == cmp_values


@dataclass(slots=True)
class RerollIfGreater(BaseReroll):
    def __str__(self) -> str:
        return f"{self.dice}r>{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice > compare

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values > cmp_values  # type: ignore


@dataclass(slots=True)
class RerollIfGreaterOrEq(BaseReroll):
    def __str__(self) -> str:
        return f"{self.dice}r>={self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice >= compare

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values >= cmp_values  # type: ignore


@dataclass(slots=True)
class RerollIfLess(BaseReroll):
    def __str__(self) -> str:
        return f"{self.dice}r<{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice < compare

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values < cmp_values  # type: ignore


@dataclass(slots=True)
class RerollIfLessOrEq(BaseReroll):
    def __str__(self) -> str:
        return f"{self.dice}r<={self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice <= compare

    @staticmethod
    def _calculate_reroll_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values <= cmp_values  # type: ignore


class Reroll:
    def __init__(self, reroll_limit: int = 1) -> None:
        self.reroll_limit = reroll_limit

    def __eq__(self, value: BaseDice | int) -> DiceModifier:  # type: ignore
        return partial(RerollEq, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)

    def __gt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(RerollIfGreater, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)

    def __ge__(self, value: BaseDice | int) -> DiceModifier:
        return partial(RerollIfGreaterOrEq, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)

    def __lt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(RerollIfLess, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)

    def __le__(self, value: BaseDice | int) -> DiceModifier:
        return partial(RerollIfLessOrEq, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)


class _RerollFactory:
    def __init__(self, dice: BaseDice, reroll_limit: int = 1) -> None:
        self.dice = dice
        self.reroll_limit = reroll_limit

    def __eq__(self, value: BaseDice | int) -> BaseReroll:  # type: ignore
        return RerollEq(dice=self.dice, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)  # type: ignore

    def __gt__(self, value: BaseDice | int) -> BaseReroll:
        return RerollIfGreater(dice=self.dice, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)  # type: ignore

    def __ge__(self, value: BaseDice | int) -> BaseReroll:
        return RerollIfGreaterOrEq(dice=self.dice, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)  # type: ignore

    def __lt__(self, value: BaseDice | int) -> BaseReroll:
        return RerollIfLess(dice=self.dice, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)  # type: ignore

    def __le__(self, value: BaseDice | int) -> BaseReroll:
        return RerollIfLessOrEq(dice=self.dice, compare=_wrap_scalar(value), reroll_limit=self.reroll_limit)  # type: ignore



from dataclasses import dataclass, field
from functools import partial
from typing import Protocol

import numpy as np
from dyce import H
from dyce.evaluation import HResult, expandable
from numpy.typing import ArrayLike

from .core import BaseDice
from .misc import DiceModifier, _wrap_scalar


@dataclass(slots=True)
class BaseExplode(BaseDice, Protocol):
    dice: BaseDice
    compare: BaseDice
    explode_depth: int = field(default=100)

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool: ...

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike: ...

    def histogram(self) -> H:
        dice_hist = self.dice.histogram()

        @expandable(sentinel=dice_hist)
        def _explode(compare: HResult, dice: HResult):
            if self._compare_histogram_outcome(dice.outcome, compare.outcome):  # type: ignore
                # Replace a high roll with sum of the roll and recursively exploding dice
                return dice.outcome + _explode(compare.h, dice.h)  # type: ignore
            else:
                return dice.outcome

        return _explode(self.compare.histogram(), dice_hist, limit=self.explode_depth - 1)  # type: ignore

    def max(self) -> int:
        return self.dice.max() * self.explode_depth

    def min(self) -> int:
        return self.dice.min()

    def generate(self, items: int) -> ArrayLike:
        results = np.zeros(items, dtype=np.int_)
        current_rolls = self.dice.generate(items)

        for _ in range(self.explode_depth + 1):
            results += current_rolls  # type: ignore
            compare_rolls = self.compare.generate(items)
            explode_mask = self._calculate_explode_mask(current_rolls, compare_rolls)

            if not np.any(explode_mask):
                break  # Exit if no dice explode in this iteration

            # Calculate how many dice need to be rerolled based on the explode_mask
            current_rolls = np.zeros(items, dtype=np.int_)
            current_rolls[explode_mask] = self.dice.generate(np.sum(explode_mask))  # type: ignore

        return results


@dataclass(slots=True)
class ExplodeEq(BaseExplode):
    def __str__(self) -> str:
        return f"{self.dice}x{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice == compare

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values == cmp_values


@dataclass(slots=True)
class ExplodeIfGreater(BaseExplode):

    def __str__(self) -> str:
        return f"{self.dice}x>{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice > compare

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values > cmp_values  # type: ignore


@dataclass(slots=True)
class ExplodeIfGreaterOrEq(BaseExplode):
    def __str__(self) -> str:
        return f"{self.dice}x>={self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice >= compare

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values >= cmp_values  # type: ignore


@dataclass(slots=True)
class ExplodeIfLess(BaseExplode):
    def __str__(self) -> str:
        return f"{self.dice}x<{self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice < compare

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values < cmp_values  # type: ignore


@dataclass(slots=True)
class ExplodeIfLessOrEq(BaseExplode):
    def __str__(self) -> str:
        return f"{self.dice}x<={self.compare}"

    @staticmethod
    def _compare_histogram_outcome(dice: int, compare: int) -> bool:
        return dice <= compare

    @staticmethod
    def _calculate_explode_mask(roll_values: ArrayLike, cmp_values: ArrayLike) -> ArrayLike:
        return roll_values <= cmp_values  # type: ignore


class Explode:
    def __init__(self, explode_depth: int = 100) -> None:
        self.explode_depth = explode_depth

    def __eq__(self, value: BaseDice | int) -> DiceModifier:  # type: ignore
        return partial(ExplodeEq, compare=_wrap_scalar(value), explode_depth=self.explode_depth)

    def __gt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(ExplodeIfGreater, compare=_wrap_scalar(value), explode_depth=self.explode_depth)

    def __ge__(self, value: BaseDice | int) -> DiceModifier:
        return partial(ExplodeIfGreaterOrEq, compare=_wrap_scalar(value), explode_depth=self.explode_depth)

    def __lt__(self, value: BaseDice | int) -> DiceModifier:
        return partial(ExplodeIfLess, compare=_wrap_scalar(value), explode_depth=self.explode_depth)

    def __le__(self, value: BaseDice | int) -> DiceModifier:
        return partial(ExplodeIfLessOrEq, compare=_wrap_scalar(value), explode_depth=self.explode_depth)


class _ExplodeFactory:
    def __init__(self, dice: BaseDice, explode_depth: int = 100) -> None:
        self.dice = dice
        self.explode_depth = explode_depth

    def __eq__(self, value: BaseDice | int) -> BaseExplode:  # type: ignore
        return ExplodeEq(self.dice, compare=_wrap_scalar(value), explode_depth=self.explode_depth)  # type: ignore

    def __gt__(self, value: BaseDice | int) -> BaseExplode:
        return ExplodeIfGreater(self.dice, compare=_wrap_scalar(value), explode_depth=self.explode_depth)  # type: ignore

    def __ge__(self, value: BaseDice | int) -> BaseExplode:
        return ExplodeIfGreaterOrEq(self.dice, compare=_wrap_scalar(value), explode_depth=self.explode_depth)  # type: ignore

    def __lt__(self, value: BaseDice | int) -> BaseExplode:
        return ExplodeIfLess(self.dice, compare=_wrap_scalar(value), explode_depth=self.explode_depth)  # type: ignore

    def __le__(self, value: BaseDice | int) -> BaseExplode:
        return ExplodeIfLessOrEq(self.dice, compare=_wrap_scalar(value), explode_depth=self.explode_depth)  # type: ignore



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np
from dyce import H, P
from dyce.evaluation import HResult, expandable
from numpy.typing import ArrayLike

from .core import BaseDice, DiceMany, Scalar


@dataclass(slots=True)
class KeepHighest(BaseDice):
    dice: BaseDice
    keep: BaseDice = field(default_factory=lambda: Scalar(1))
    of: BaseDice = field(default_factory=lambda: Scalar(1))

    def __post_init__(self):
        if isinstance(self.dice, DiceMany):
            self.of = self.dice.total
            self.dice = self.dice.dice
        if isinstance(self.keep, int):
            self.keep = Scalar(self.keep)

    def histogram(self) -> H:
        @expandable
        def kh(dice: HResult, keep: HResult, of: HResult):
            return (of.outcome @ P(dice.h)).h(slice(-keep.outcome, None))  # type: ignore

        return kh(self.dice.histogram(), self.keep.histogram(), self.of.histogram())

    def __str__(self) -> str:
        keep = str(self.keep)
        if isinstance(self.keep, Scalar) and self.keep.value == 1:
            keep = ""
        return f"{self.of}{self.dice}kh{keep}"

    def max(self) -> int:
        # Assuming the best rolls are kept, the maximum is the highest dice value times the maximum number of keeps,
        # but it cannot exceed the total number of dice rolled.
        return self.dice.max() * min(self.keep.max(), self.of.max())

    def min(self) -> int:
        # If keeping less than or equal to the number of dice rolled, the minimum would still be the dice's minimum value
        # times the minimum number of keeps, as we're assuming the least favorable (lowest) high rolls are kept.
        return self.dice.min() * self.keep.min()

    def generate(self, items: int) -> ArrayLike:
        of_rolls = self.of.generate(items)
        dice_rolls = self.dice.generate(np.sum(of_rolls))
        keep_rolls = self.keep.generate(items)
        results = np.empty(items, dtype=np.int_)

        start_idx = 0
        for i in range(items):
            num_rolls = of_rolls[i]  # type: ignore
            keep = min(keep_rolls[i], num_rolls)  # type: ignore
            results[i] = np.sum(np.sort(dice_rolls[start_idx : start_idx + num_rolls])[-keep:])  # type: ignore
            start_idx += num_rolls  # type: ignore

        return results


@dataclass(slots=True)
class KeepLowest(BaseDice):
    dice: BaseDice
    keep: BaseDice = field(default_factory=lambda: Scalar(1))
    of: BaseDice = field(default_factory=lambda: Scalar(1))

    def __post_init__(self):
        if isinstance(self.dice, DiceMany):
            self.of = self.dice.total
            self.dice = self.dice.dice
        if isinstance(self.keep, int):
            self.keep = Scalar(self.keep)

    def histogram(self) -> H:
        @expandable
        def kl(dice: HResult, keep: HResult, of: HResult):
            return (of.outcome @ P(dice.h)).h(slice(None, keep.outcome))  # type: ignore

        return kl(self.dice.histogram(), self.keep.histogram(), self.of.histogram())

    def __str__(self) -> str:
        keep = str(self.keep)
        if isinstance(self.keep, Scalar) and self.keep.value == 1:
            keep = ""
        return f"{self.of}{self.dice}kl{keep}"

    def max(self) -> int:
        # For keeping the lowest, the max is now the dice's max value times the minimum number of keeps,
        # because even when keeping the lowest, the max scenario would involve the highest possible "low" values.
        return self.dice.max() * min(self.keep.max(), self.of.max())

    def min(self) -> int:
        # The minimum is the dice's minimum value times the number of keeps, assuming the lowest possible outcomes are kept.
        return self.dice.min() * self.keep.min()

    def generate(self, items: int) -> ArrayLike:
        of_rolls = self.of.generate(items)
        dice_rolls = self.dice.generate(np.sum(of_rolls))
        keep_rolls = self.keep.generate(items)
        results = np.empty(items, dtype=np.int_)

        start_idx = 0
        for i in range(items):
            num_rolls = of_rolls[i]  # type: ignore
            keep = min(keep_rolls[i], num_rolls)  # type: ignore
            results[i] = np.sum(np.sort(dice_rolls[start_idx : start_idx + num_rolls])[:keep])  # type: ignore
            start_idx += num_rolls  # type: ignore

        return results


@dataclass(slots=True)
class DropHighest(BaseDice):
    dice: BaseDice
    drop: BaseDice = field(default_factory=lambda: Scalar(1))
    of: BaseDice = field(default_factory=lambda: Scalar(1))

    def __post_init__(self):
        if isinstance(self.dice, DiceMany):
            self.of = self.dice.total
            self.dice = self.dice.dice
        if isinstance(self.drop, int):
            self.drop = Scalar(self.drop)

    def histogram(self) -> H:
        @expandable
        def dh(dice: HResult, drop: HResult, of: HResult):
            return (of.outcome @ P(dice.h)).h(slice(None, drop.outcome))  # type: ignore

        return dh(self.dice.histogram(), self.drop.histogram(), self.of.histogram())

    def __str__(self) -> str:
        drop = str(self.drop)
        if isinstance(self.drop, Scalar) and self.drop.value == 1:
            drop = ""
        return f"{self.of}{self.dice}dh{drop}"

    def max(self) -> int:
        # Maximum possible value after dropping the highest rolls
        return self.dice.max() * (self.of.max() - self.drop.min())

    def min(self) -> int:
        # Minimum possible value after dropping the highest rolls
        return self.dice.min() * max(0, self.of.min() - self.drop.max())

    def generate(self, items: int) -> np.ndarray:
        of_rolls = self.of.generate(items)
        drop_rolls = self.drop.generate(items)
        dice_rolls = self.dice.generate(np.sum(of_rolls))

        # Initialize results array
        results = np.empty(items, dtype=np.int_)

        start_idx = 0
        for i in range(items):
            num_rolls = of_rolls[i]  # type: ignore
            drop = min(num_rolls, drop_rolls[i])  # type: ignore

            sorted_rolls = np.sort(dice_rolls[start_idx : start_idx + num_rolls])  # type: ignore
            results[i] = np.sum(sorted_rolls[:-drop] if drop > 0 else sorted_rolls)  # type: ignore

            start_idx += num_rolls  # type: ignore

        return results


@dataclass(slots=True)
class DropLowest(BaseDice):
    dice: BaseDice
    drop: BaseDice = field(default_factory=lambda: Scalar(1))
    of: BaseDice = field(default_factory=lambda: Scalar(2))  # Adjusted default for consistency

    def __post_init__(self):
        if isinstance(self.dice, DiceMany):
            self.of = self.dice.total
            self.dice = self.dice.dice
        if isinstance(self.drop, int):
            self.drop = Scalar(self.drop)

    def histogram(self) -> H:
        @expandable
        def dh(dice: HResult, drop: HResult, of: HResult):
            return (of.outcome @ P(dice.h)).h(slice(-drop.outcome, None))  # type: ignore

        return dh(self.dice.histogram(), self.drop.histogram(), self.of.histogram())

    def __str__(self) -> str:
        drop = str(self.drop)
        if isinstance(self.drop, Scalar) and self.drop.value == 1:
            drop = ""
        return f"{self.of}{self.dice}dl{drop}"

    def max(self) -> int:
        # Maximum possible value after dropping the lowest rolls.
        # The max calculation now considers the possibility of dropping less valuable rolls.
        return self.dice.max() * (self.of.max() - self.drop.min())

    def min(self) -> int:
        # Minimum possible value after dropping the lowest rolls.
        # Adjusted to consider the effect of dropping the lowest possible rolls.
        return self.dice.min() * max(0, self.of.min() - self.drop.max())

    def generate(self, items: int) -> np.ndarray:
        of_rolls = self.of.generate(items)
        drop_rolls = self.drop.generate(items)
        dice_rolls = self.dice.generate(np.sum(of_rolls))

        results = np.empty(items, dtype=np.int_)

        start_idx = 0
        for i in range(items):
            num_rolls = of_rolls[i]  # type: ignore
            drop = min(num_rolls, drop_rolls[i])  # type: ignore

            sorted_rolls = np.sort(dice_rolls[start_idx : start_idx + num_rolls])  # type: ignore
            results[i] = np.sum(sorted_rolls[drop:] if drop > 0 else sorted_rolls)  # type: ignore

            start_idx += num_rolls  # type: ignore

        return results



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

from numpy.typing import ArrayLike

from .core import BaseDice


@dataclass(slots=True)
class WithRollCallback(BaseDice):
    dice: BaseDice
    roll_callback: Callable[[int], None]

    def __str__(self) -> str:
        return self.dice.__str__()

    def max(self) -> int:
        return self.dice.max()

    def min(self) -> int:
        return self.dice.min()

    def roll(self) -> int:
        result = self.dice.roll()
        if self.roll_callback is not None:
            self.roll_callback(result)
        return result

    def generate(self, items: int) -> ArrayLike:
        return self.dice.generate(items)


@dataclass(slots=True)
class WithGenerateCallback(BaseDice):
    dice: BaseDice
    generate_callback: Callable[[ArrayLike], None]

    def __str__(self) -> str:
        return self.dice.__str__()

    def max(self) -> int:
        return self.dice.max()

    def min(self) -> int:
        return self.dice.min()

    def generate(self, items: int) -> ArrayLike:
        result = self.dice.generate(items)
        if self.generate_callback is not None:
            self.generate_callback(result)
        return result



import numpy as np


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class Rng(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self._rng: np.random.Generator = None  # type: ignore

    @property
    def rng(self) -> np.random.Generator:
        if self._rng is None:
            raise RuntimeError("No RNG generator ")
        return self._rng

    def set_rng(self, rng: np.random.Generator):
        self._rng = rng


Rng().set_rng(np.random.default_rng())



from . import random
from .callback import WithGenerateCallback, WithRollCallback
from .compare import Ge, Gt, Le, Lt, Limit
from .core import (
    BaseDice,
    Dice,
    DiceMany,
    RangeDice,
    Scalar,
)
from .explode import Explode
from .math import (
    DiceAdd,
    DiceDiv,
    DiceMul,
    DiceSub,
)
from .reroll import Reroll
from .transformations import DropHighest, DropLowest, KeepHighest, KeepLowest

s = Scalar
d = Dice
kh = KeepHighest
kl = KeepLowest
dh = DropHighest
dl = DropLowest
x = Explode
r = Reroll
rng = RangeDice
lim = Limit
